import pytest

from .assets import get_assets
from .tester import Tester


# User can specify these options through execution flags
def pytest_addoption(parser):
    """Add options for contract assets file handling"""

    group = parser.getgroup('ethereum', 'ethereum testing support')
    group.addoption('--assets-file', action='store', default=None, metavar='path',
            help='assets file for coverage, default: none')
    # TODO: # of Accounts must be set at startup


# Load assets file into memory
@pytest.fixture(scope='session')
def assets(pytestconfig):
    assets = {}
    assets_file = pytestconfig.option.assets_file
    if assets_file:
        assets = get_assets(assets_file)
    return assets


# Fixture is initialized from default for every test
@pytest.fixture
def tester(assets):
    return Tester(assets)

# alias fixture
@pytest.fixture
def t(tester):
    return tester



import pytest

from web3 import Web3
from web3.providers.eth_tester import EthereumTesterProvider

from eth_tester import EthereumTester
from eth_tester.exceptions import TransactionFailed

from .account import Account
from .contract import ContractFactory


class Tester:
    def __init__(self, compiled_interfaces):
        self.__t = EthereumTester()
        self.__w3 = Web3(EthereumTesterProvider(self.__t))
        # Empty by default, but can be set on initialization
        self.__compiled_interfaces = compiled_interfaces
        
        # TODO Set starting balance for all accounts to something very high
        #      and blk gas limit for chain to something approaching average

    def contracts(self, name):
        if ':' not in name:
            # If you don't specify which contract in file, use filebase
            name += ':{}'.format(name.split('/')[-1].split('.')[0])
        interface = self.__compiled_interfaces[name]
        return self.new_contract(interface)

    # alias
    def c(self, name):
        return self.contracts(name)

    def new_contract(self, interface):
        return ContractFactory(self.__w3, interface)
        
    @property
    def accounts(self):
        return [Account(self.__w3, a) for a in self.__t.get_accounts()]

    # Alias
    @property
    def a(self):
        return self.accounts
    
    @property
    def tx_fails(self):
        return pytest.raises(TransactionFailed)

    def mine_blocks(self, number=1):
        self.__t.mine_blocks(number)

    def now(self):
        # TODO Get this from the Ethereum block timestamp
        return self.__w3.eth.getBlock('latest')['timestamp']



from web3 import Web3


class Account(str):
    def __new__(cls, w3: Web3, address):
        obj = super().__new__(cls, address)
        obj._w3 = w3
        obj._address = address
        return obj

    # Send Ether
    def send(self, address, amount):
        self._w3.eth.sendTransaction({'to': address, 'from': self._address, 'value': amount})

    @property
    def balance(self):
        return self._w3.eth.getBalance(self._address)



from web3 import Web3
from web3.contract import ImplicitContract

from .log import Log
from .utils import (
        clean_modifiers,
        get_event_processors,
        get_event_signatures,
    )


class ContractInstance:
    """Deployed instance of a contract"""

    def __init__(self, w3: Web3, address, interface):
        self.__w3 = w3
        self.__address = address
        self.__instance = self.__w3.eth.contract(self.__address, **interface,
                ContractFactoryClass=ImplicitContract)
        # Register new filter to watch for logs from this instance's address
        self.__filter = self.__w3.eth.filter({
                # Include events from the deployment stage
                'fromBlock': self.__w3.eth.blockNumber - 1,
                'address': self.__address
            })
        
        self.__event_signatures = get_event_signatures(interface['abi'])
        self.__event_processors = get_event_processors(interface['abi'])

    def __getattr__(self, name):
        """Delegates to either specialized methods or instance ABI"""
        if name in dir(self):
            # Specialized testing methods
            return getattr(self, name)
        elif name in self._events:
            return self._gen_log(name)
        else:
            # Method call of contract instance
            return getattr(self.__instance, name)

    @property
    def _events(self):
        return self.__event_signatures.keys()

    def _gen_log(self, name):
        return lambda v: Log(name, v)

    @property
    def address(self):
        """This contract's address"""
        return self.__address

    @property
    def balance(self):
        """Ether balance of this contract (in wei)"""
        return self.__w3.eth.getBalance(self.__address)

    @property
    def codesize(self):
        """Codesize of this contract (in bytes)"""
        return len(self.__w3.eth.getCode(self.__address)[2:])/2

    @property
    def hascode(self):
        """Check if this contract currently has code (usually indicating suicide)"""
        return self.codesize != 0

    def _process_logs(self, logs):
        processed_logs = []
        for log in logs:
            log_signature = log['topics'][0]
            if log_signature in self.__event_processors.keys():
                p_log = self.__event_processors[log_signature](log)
                processed_logs.append(Log(p_log['event'], p_log['args']))
        return processed_logs

    @property
    def logs(self):
        """Returns all the event logs ever added for this contract"""
        return self._process_logs(self.__filter.get_all_entries())


class ContractFactory:
    """Factory (prototype) of a contract"""
    def __init__(self, w3: Web3, interface):
        self.__w3 = w3
        self.__interface = interface
        self.__contract_factory = self.__w3.eth.contract(**self.__interface)

    def __call__(self, *args, **kwargs):
        """Deploy a new instance of this contract"""

        #TODO: HACK, awaiting resolution of web3.py/#666
        if 'transact' in kwargs.keys():
            kwargs['transaction'] = kwargs['transact']
            del kwargs['transact']

        # Our encapsulating classes need to be evaluated here
        # in order to avoid bugs when web3 processes our modifiers
        kwargs = clean_modifiers(kwargs)

        tx_hash = self.__contract_factory.constructor(*args).transact(**kwargs)
        address = self.__w3.eth.getTransactionReceipt(tx_hash)['contractAddress']
        return ContractInstance(self.__w3, address, self.__interface)



from collections import Mapping

class Log(Mapping):
    def __new__(cls, event, args):
        obj = super().__new__(cls)
        obj._event = event
        obj._args = args
        return obj

    def __eq__(self, other):
        if not isinstance(other, Log):
            return False
        if self._event != other._event:
            return False
        return self._args == other._args

    def __iter__(self):
        return iter(self._args)

    def __len__(self):
        return len(self._args)

    def __getitem__(self, key):
        return self._args[key]



from setuptools import setup

setup(
    name="pytest-ethereum",
    version='0.1.0a14',
    description='py.test plugin for testing Ethereum smart contracts',
    long_description='TODO',
    license='MIT',
    author='Bryant Eisenbach',
    author_email='bryant@dappdevs.org',
    url='https://github.com/fubuloubu/pytest-ethereum',
    python_requires='>=3.6',
    packages = ['pytest_ethereum'],
    install_requires=[
        'pytest',
        'eth-tester[pyethereum21]',  # eventually switch to py-evm backend
        'web3>=4.2.0',
    ],

    # the following makes a plugin available to pytest
    entry_points = {
        'pytest11': [
            'name_of_plugin = pytest_ethereum.plugin',
        ]
    },

    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3.6',
        # custom PyPI classifier for pytest plugins
        "Framework :: Pytest",
    ],
)



from eth_utils import event_abi_to_log_topic
from web3.utils.events import get_event_data
from functools import partial as partial_fn


def clean_modifiers(modifiers):
    cleaned_modifiers = modifiers.copy()
    for name, modifier in modifiers.items():
        for key, value in modifier.items():
            if not isinstance(value, str) or not isinstance(value, int):
                cleaned_modifiers[name][key] = str(value)
    return cleaned_modifiers


def get_event_signatures(abi_list):
    signatures = dict()
    for abi in abi_list:
        if abi['type'] == 'event':
            signatures[abi['name']] = event_abi_to_log_topic(abi)
    return signatures


def get_event_processors(abi_list):
    processors = dict()
    for abi in abi_list:
        if abi['type'] == 'event':
            processors[event_abi_to_log_topic(abi)] = partial_fn(get_event_data, abi)
    return processors
